import std.stdio;
import std.file;
import std.ctype;
import std.string;
import std.c.stdlib;
import std.format;
import std.path;
import std.stream;

import tostring;
import util;

void die(string m) {
    writefln(m);
    exit(1);
}

void check(B)(B b, string m) {
    if (!b) die(m);
}

class O {
    abstract string dump();
    void print() {
        writefln(dump);
    }
}

class T : O {
    void emit() {
        .emit(format("Thunk!(%s) ", dtype));
    }
    void emit_nothunk() {
        .emit(format("%s ", dtype));
    }
    abstract string dtype();
}

class TP : T {
    enum P {
        INT, STR, BOOL,
        IO,  // TODO
    }
    this (P p0) {
        p = p0;
    }
    override string dump() {
        return ["INT", "STR", "BOOL", "IO"][cast(int)p];
    }
    override string dtype() {
        return ["int", "string", "bool", "IO"][cast(int)p];
    }

    static TP STR;
    static TP INT;
    static TP BOOL;
    static TP IO;
    static this() {
        STR = new TP(P.STR);
        INT = new TP(P.INT);
        BOOL = new TP(P.BOOL);
        IO = new TP(P.IO);
    }

    P p;
}

class S : O {
    abstract string dump_();
    override string dump() {
        string s = dump_;
        if (t) {
            s ~= "[" ~ t.dump ~ "]";
        }
        return s;
    }
    T t;
}

class SS : S {
    string s;
    //mixin ctor;
    this(string s0) {
        s = s0;
        t = TP.STR;
    }

    override string dump_() {
        return `"` ~ s ~ `"`;
    }
}

class SI : S {
    string s;
    //mixin ctor;
    this(string s0) {
        s = s0;
    }

    override string dump_() {
        return s;
    }
}

class SN : S {
    int n;
    //mixin ctor;
    this(int n0) {
        n = n0;
        t = TP.INT;
    }

    override string dump_() {
        return tostr(n);
    }
}

class SL : S {
    S[] ss;
    //mixin ctor;
    this(S[] ss0) {
        ss = ss0;
    }

    override string dump_() {
        string[] s;
        foreach (c; ss) {
            s ~= c.dump;
        }
        return "(" ~ join(s, " ") ~ ")";
    }

    S opIndex(int i) {
        return (i < ss.length) ? ss[i] : null;
    }
}

class FuncDecl {
    this(T r, T[] a) {
        ret = r;
        args = a;
    }
    T ret;
    T[] args;
}

FuncDecl[string] functbl;
File of;

S parse(string f, inout int p) {
first:
    switch (f[p]) {
    case ' ':
    case '\n':
    case '\r':
        p++;
        goto first;
    case '(': {
        p++;
        S[] ss;
        while (f[p] != ')') {
            ss ~= parse(f, p);
            while (isspace(f[p])) p++;
        }
        p++;
        return new SL(ss);
    }
    case '"': {
        int first = ++p;
        char[] s;
        while (f[p] != '"') {
            p++;
        }
        p++;
        return new SS(f[first..p-1]);
    }
    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9': {
        int n = 0;
        while (std.ctype.isdigit(f[p])) {
            n = n * 10 + f[p]-'0';
            p++;
        }
        return new SN(n);
    }
    default: {
        char[] s;
        int first = p;
        while (!isspace(f[p]) && f[p] != ')' && f[p] != '(') {
            p++;
        }
        return new SI(f[first..p]);
    }
    }
}

template toS(B, string n) {
    B toS(S s) {
        check(s, "expected " ~ n);
        B l = cast(B)s;
        check(l, "expected " ~ n);
        return l;
    }
}

alias toS!(SL, "sexpr") toSL;
alias toS!(SI, "identifier") toSI;
alias toS!(SS, "string") toSS;
alias toS!(SN, "number") toSN;

bool isS(B)(S s) {
    B b = cast(B)s;
    return b !is null;
}

alias isS!(SL) isSL;
alias isS!(SI) isSI;
alias isS!(SS) isSS;
alias isS!(SN) isSN;

void submit_type(S s, T t, inout T[string] vars) {
    s.t = t;
    if (isSI(s)) {
        SI id = toSI(s);
        if (id.s in vars) {
            vars[id.s] = t;
        }
        else {
            die("undefined variable: " ~ id.s);
        }
    }
}

void typecheck_l(SL st, inout T[string] vars) {
    string n = toSI(st[0]).s;
    if (n == "let") {
        string vn = toSI(st[1]).s;
        typecheck_s(st[2], vars);
        st[1].t = vars[vn] = st[2].t;
        typecheck_s(st[3], vars);
        st.t = st[3].t;
        vars.remove(vn);
    }
    else if (n == "if") {
        typecheck_s(st[1], vars);
        check(st[1].t == TP.BOOL, "non bool condition for if");
        typecheck_s(st[2], vars);
        typecheck_s(st[3], vars);
        if (!st[2].t) {
            check(st[3].t, "failed type reduction");
            submit_type(st[2], st[3].t, vars);
        }
        else if (!st[3].t) {
            submit_type(st[3], st[2].t, vars);
        }
        check(st[2].t == st[3].t,
              format("type mismatch in if %s vs %s", st[2].dump, st[3].dump));
        st.t = st[2].t;
    }
    else {
        writefln(n);
        check(n in functbl, "undefined function: " ~ n);
        FuncDecl decl = functbl[n];
        int i = 1;
        foreach (inout argtype; decl.args) {
            S arg = st[i];
            if (!arg) break;
            typecheck_s(arg, vars);
            if (!argtype) {
                if (arg.t) {
                    argtype = arg.t;
                }
            }
            else if (arg.t) {
                check(arg.t is argtype,
                      format("type of argument %d of %s is %s but comes %s",
                             i, n, argtype.dump, arg.t.dump));
            }
            else {
                //writefln(toSI(arg).s);
                submit_type(arg, argtype, vars);
            }
            i++;
        }
        st.t = decl.ret;
    }
}

void typecheck_i(SI ident, inout T[string] vars) {
    check(ident.s in vars, "undefined variable: " ~ ident.s);
    ident.t = vars[ident.s];
}

void typecheck_s(S st, inout T[string] vars) {
    writefln("TYPECHECK ", st.dump);
    if (isSL(st)) {
        typecheck_l(toSL(st), vars);
    }
    else if (isSI(st)) {
        typecheck_i(toSI(st), vars);
    }
    //writefln("END TYPECHECK ", st.dump);
}

void typecheck(SL func) {
    SI n = toSI(func[0]);
    check(n.s == "funcdef", "expected funcdef");

    SL args = toSL(func[2]);
    T[string] vars;
    foreach (arg; args.ss) {
        vars[toSI(arg).s] = null;
    }
    T[] argtypes;
    foreach (arg; args.ss) {
        arg.t = vars[toSI(arg).s];
        argtypes ~= arg.t;
    }
    functbl[toSI(func[1]).s] = new FuncDecl(null, argtypes);

    typecheck_s(func[3], vars);
    func.t = func[3].t;
    // TODO: ugly
    argtypes = [];
    foreach (arg; args.ss) {
        arg.t = vars[toSI(arg).s];
        argtypes ~= arg.t;
    }
    functbl[toSI(func[1]).s] = new FuncDecl(func[3].t, argtypes);
}

void emit(string s) {
    of.writeString(s);
}

void emit_ident(S s) {
    emit(toSI(s).s);
}

string mangle(string n) {
    string r = "_";
    foreach (c; n) {
        if (isalnum(c)) r ~= c;
        else {
            int i = cast(int)c;
            r ~= format("__%d", i);
        }
    }
    return r;
}

void emit_l(SL st) {
    string n = toSI(st[0]).s;
    if (n == "let") {
        emit("((");
        st[1].t.emit();
        emit_ident(toSI(st[1]));
        emit(")");
        emit(" { ");
        emit("return ");
        emit_s(st[3]);
        emit("; })(");
        emit_s(st[2]);
        emit(")");
        //emit(";\n");
    }
    else if (n == "if") {
        emit_s(st[1]);
        emit("()");
        emit(" ? ");
        emit_s(st[2]);
        emit(" : ");
        emit_s(st[3]);
    }
    else {
        check(n in functbl, "internal error: " ~ n);
        emit("{ return ");
        emit(mangle(n));
        emit("(");
        for (int i = 1;; i++) {
            S arg = st[i];
            if (!arg) break;
            if (i > 1) emit(", ");
            //arg.t.emit();
            emit_s(arg);
        }
        emit(")");
        emit(";}");
    }
}

void emit_s(S s) {
    if (isSI(s)) {
        emit_ident(s);
    }
    else if (isSS(s)) {
        //emit(format("\"%s\"", toSS(s).s));
        emit("new Thunk!(string)(\"");
        emit(toSS(s).s);
        emit("\")");
    }
    else {
        emit("thunk(");
        if (isSL(s)) {
            emit_l(toSL(s));
        }
        else if (isSN(s)) {
            emit(format("%d", toSN(s).n));
            //emit(format("new Thunk!(int)(%d)", toSN(s).n));
        }
        emit(")");
    }
}

void emit_func(SL func) {
    emit("extern(C) ");

    SL args = toSL(func[2]);
    int poly_num = 0;
    foreach (arg; args.ss) {
        if (!toSI(arg).t) {
            poly_num++;
        }
    }

    if (func.t && !poly_num) {
        func.t.emit_nothunk();
    }
    else {
        // TODO
        emit(format("template %s(", mangle(toSI(func[1]).s)));
        for (int i = 0; i < poly_num; i++) {
            if (i) emit(", ");
            emit(format("T%d", i));
        }
        emit(") {\n");
        if (!func.t) {
            emit("T0 ");
        }
    }

    emit(mangle(toSI(func[1]).s));
    emit("(");
    bool first = true;
    int n = 0;
    foreach (arg; args.ss) {
        if (!first) emit(",");
        if (toSI(arg).t) {
            toSI(arg).t.emit();
        }
        else {
            emit(format("Thunk!(T%d) ", n));
            n++;
        }
        emit_ident(toSI(arg));
        first = false;
    }
    emit(") {\n");
    emit(" return ");
    //emit(" return new ");
    //func.t.emit();
    //emit("({return ");
    emit_s(func[3]);
    //emit(";})");
    emit("();\n");
    emit("}\n");
    if (!func.t || poly_num) {
        emit("}\n");
    }
}

void init() {
    functbl[`++`] = new FuncDecl(TP.STR, [TP.STR, TP.STR]);
    functbl[`putStr`] = new FuncDecl(TP.IO, [TP.STR]);
    functbl[`putInt`] = new FuncDecl(TP.IO, [TP.INT]);
    functbl[`return`] = new FuncDecl(TP.IO, []);
    functbl[`mod`] = new FuncDecl(TP.INT, [TP.INT, TP.INT]);
    functbl[`==`] = new FuncDecl(TP.BOOL, [TP.INT, TP.INT]);  // TODO
    functbl[`!=`] = new FuncDecl(TP.BOOL, [TP.INT, TP.INT]);  // TODO
    functbl[`<=`] = new FuncDecl(TP.BOOL, [TP.INT, TP.INT]);  // TODO
    functbl[`>>`] = new FuncDecl(TP.IO, [TP.IO, TP.IO]);  // TODO
    functbl[`+`] = new FuncDecl(TP.INT, [TP.INT, TP.INT]);
    functbl[`-`] = new FuncDecl(TP.INT, [TP.INT, TP.INT]);
}

int main(string[] args) {
    if (args.length < 2) {
        writefln("Usage: %s file", args[0]);
        return 1;
    }

    init();

    string src = args[1];

    int pos = 0;
    S sexp = parse(cast(string)std.file.read(src), pos);
    writefln(sexp.dump);

    writefln("=== type check ===");
    foreach (s; toSL(sexp).ss) {
        writefln(s.dump);
        typecheck(toSL(s));
        writefln(s.dump);
    }

    writefln("=== code generation ===");
    string dsrc = basename(src, ".hss") ~ ".d";
    of = new File;
    of.create(dsrc);
    emit("import rt;\n");
    foreach (s; toSL(sexp).ss) {
        emit_func(toSL(s));
    }

    return 0;
}
